#ifndef TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED

#include <bit>
#include <cstdint>
#include <type_traits>


namespace typon::fdt
{

  template <typename T>
  struct ring_buffer
  {
    using u8 = std::uint_least8_t;
    using u64 = std::uint_fast64_t;

    const u64 _mask;
    ring_buffer * _next;
    T * const _array;

    ring_buffer(u8 bits, ring_buffer * next = nullptr) noexcept
      : _mask((u64(1) << bits) - 1)
      , _next(next)
      , _array(new T[this->capacity()])
    {}

    ~ring_buffer()
    {
      delete [] _array;
      if (_next)
      {
        delete _next;
      }
    }

    u64 capacity() noexcept
    {
      return _mask + 1;
    }

    void put(u64 index, T object) noexcept
    {
      _array[index & _mask] = std::move(object);
    }

    T get(u64 index) noexcept
    {
      return _array[index & _mask];
    }

    ring_buffer * fill(ring_buffer * sink, u64 start, u64 end) noexcept
    {
      for (u64 i = start; i < end; i++)
      {
        sink->put(i, get(i));
      }
      return sink;
    }

    ring_buffer * grow(u64 start, u64 end) noexcept
    {
      auto buffer = new ring_buffer(std::countr_one(_mask) + 1, this);
      return fill(buffer, start, end);
    }

    ring_buffer * shrink(u64 start, u64 end) noexcept
    {
      ring_buffer * last = nullptr;
      auto next = this;
      auto size = (end - start);
      auto threshold = size * 2;
      while (next->_next && next->_next->capacity() >= threshold)
      {
        last = next;
        next = next->_next;
      }
      if (!last)
      {
        return nullptr;
      }
      return fill(std::exchange(last->_next, nullptr), start, end);
    }
  };

}


#endif //  TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED
